package	jp.co.sds_corp.report;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.text.DecimalFormat;
import java.util.Date;
import java.util.Hashtable;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;

import jp.co.sds_corp.report.source.SourceFactory;
import jp.co.sds_corp.util.common.CmnDate;
import jp.co.sds_corp.util.common.WarekiFormat;
import jp.co.sds_corp.util.config.DOMConfiguration;
import jp.co.sds_corp.util.config.DOMWriter;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;
import jp.co.sds_corp.util.script.ScriptParserFactory;

import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.MimeConstants;
import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFDataFormat;
import org.apache.poi.hssf.usermodel.HSSFFont;
import org.apache.poi.hssf.usermodel.HSSFPrintSetup;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.CellRangeAddress;
import org.apache.poi.hssf.util.HSSFColor;

/**
<a href="http://jakarta.apache.org/poi/">Jakarta POI</a>,
<a href="http://xml.apache.org/fop/">Apache FOP</a>
を使用して、Excel帳票をPDFに出力します。
<pre>
&lt;Report
	name="ReportTest"
	type="PDF"
	template="${Report.dir}/table.xls"
	sheet="Sheet1"
	fofile="result.fo"
	debug="true"&gt;
	&lt;Source ... /&gt;
	&lt;Script ... /&gt;
&lt;/Report&gt;
</pre>
Source要素については、{@link SourceFactory}を参照してください。
Script要素については、{@link ScriptParserFactory}を参照してください。
*/
public class ExcelPDFReport extends ExcelReport
{
	CmnLogger logger = CmnLogFactory.getInstance().getLogger(this);
	static long dayDiff;
	static {
		try {
			CmnDate dtFr = new CmnDate( "19000101" );
			CmnDate dtTo = new CmnDate( "19700101" );
			dayDiff = dtTo.diffDays( dtFr ) + 2;
		} catch ( Exception ex ) {
		}
	}
	String	userconfig;
	String	fofile;
	String	fontfamily;

	/**
	* 設定情報で初期化します。
	* @param    config  設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		fofile = config.getAttributes().get( "fofile", null );
		fontfamily = config.getAttributes().get( "fontfamily", "KochiGothic" );
		userconfig = config.getAttributes().get( "userconfig", null );
		if ( userconfig != null ) {
			if ( getProperty() != null ) {
				userconfig = getProperty().replaceValue( userconfig );
			}
			if ( isDebug()) {
				logger.debug( "userconfig=[" + userconfig + "]" );
			}
		}
	}

	/**
	* 帳票を出力します。
	* @param	out		出力先
	*/
	public void print( OutputStream out ) throws Exception {
		if ( !bPrint ) {
			print();
		}
		HSSFPrintSetup prset = result.getSheetAt( 0 ).getPrintSetup();
		try {
			DOMConfiguration root = DOMConfiguration.createDocument( "fo:root" );
			IConfiguration master = initPrint( root );
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			for ( int i = 0; i < result.getNumberOfSheets(); i++ ) {
				HSSFSheet sheet = result.getSheetAt( i );
				if ( isDebug()) {
					logger.debug( "print " + result.getSheetName( i ));
				}
				printSheet( root, master, result, i, sheet );
			}
			bos.close();

			StringWriter buf = new StringWriter();
			DOMWriter wr = new DOMWriter( buf );
			wr.write( root, "fo:root" );
			buf.close();

			if ( fofile != null ) {
				writeFo( buf.toString() );
			}

			FopFactory fopFactory = FopFactory.newInstance();
			if ( userconfig != null ) {
				fopFactory.setUserConfig( new File( userconfig ));
			}
			Fop fop = fopFactory.newFop( MimeConstants.MIME_PDF, out );
			TransformerFactory factory = TransformerFactory.newInstance();
			Transformer transformer = factory.newTransformer();
			Source foFile = new StreamSource( new StringReader( buf.toString() ));
			Result res = new SAXResult(fop.getDefaultHandler());
			transformer.transform(foFile, res);
			out.flush();
		} catch ( Exception ex ) {
			ex.printStackTrace();
		}
	}

	protected void writeFo( String buf ) throws Exception {
		FileWriter wr = new FileWriter( fofile );
		wr.write( buf );
		wr.flush();
		wr.close();
	}

	IConfiguration initPrint( IConfiguration root )
		throws Exception
	{
		root.setAttribute( "xmlns:fo", "http://www.w3.org/1999/XSL/Format" );
		IConfiguration master = root.addChild( "fo:layout-master-set" );
		return master;
	}

	double	width;
	short	scale;

	void printSheet( IConfiguration root, IConfiguration master,
						HSSFWorkbook book, int nPage, HSSFSheet sheet )
		throws Exception
	{
		HSSFDataFormat format = book.createDataFormat();

		String ref = "page" + nPage;
		IConfiguration layout
			= addPageMaster( master, ref, sheet.getPrintSetup() );
		String sWidth = layout.getAttributes().get( "page-width" );
		sWidth = sWidth.substring( 0, sWidth.length() - 2 );
		width = Double.valueOf( sWidth ).doubleValue();

		IConfiguration page = root.addChild( "fo:page-sequence" );
		page.setAttribute( "master-reference", ref );
		IConfiguration flow = page.addChild( "fo:flow" );
		flow.setAttribute( "flow-name", "xsl-region-body" );

		IConfiguration table = flow.addChild( "fo:table" );
		table.setAttribute( "table-layout", "fixed" );
		table.setAttribute( "width", width + "cm" );
		int nMaxCol = 0;
		for ( int i = 0; i <= sheet.getLastRowNum(); i++ ) {
			HSSFRow row = sheet.getRow( i );
			if ( row == null ) {
				continue;
			}
			if ( nMaxCol < row.getLastCellNum() ) {
				nMaxCol = row.getLastCellNum();
			}
		}
		double all = 0.0;
		for ( int i = 0; i < nMaxCol; i++ ) {
			all += (double )sheet.getColumnWidth( i );
		}
		for ( int i = 0; i < nMaxCol; i++ ) {
			IConfiguration c1 = table.addChild( "fo:table-column" );
			double w = (double )sheet.getColumnWidth( i );
			c1.setAttribute( "column-width",
				String.valueOf( w * 100 / all ) + "%" );
		}
		byte[][] pos = new byte[sheet.getLastRowNum()+1][nMaxCol];
		byte[][] bborder = new byte[sheet.getLastRowNum()+1][nMaxCol];
		byte[][] rborder = new byte[sheet.getLastRowNum()+1][nMaxCol];
		for ( int i = 0; i < sheet.getNumMergedRegions(); i++ ) {
			CellRangeAddress reg = sheet.getMergedRegion( i );
			for ( int j = reg.getFirstRow(); j <= reg.getLastRow(); j++ ) {
				for ( int k = reg.getFirstColumn(); k <= reg.getLastColumn(); k++ ) {
					pos[j][k] = -1;
					HSSFCellStyle hstyle = sheet.getRow( j )
										.getCell( k ).getCellStyle();
					if ( hstyle.getBorderBottom() != HSSFCellStyle.BORDER_NONE )
					{
						bborder[j][k] = 1;
					}
					if ( hstyle.getBorderRight() != HSSFCellStyle.BORDER_NONE )
					{
						rborder[j][k] = 1;
					}
				}
			}
			pos[reg.getFirstRow()][reg.getFirstColumn()] = 1;
		}

		IConfiguration body = table.addChild( "fo:table-body" );
		for ( int i = 0; i <= sheet.getLastRowNum(); i++ ) {
			IConfiguration crow = body.addChild( "fo:table-row" );
			HSSFRow row = sheet.getRow( i );
			if ( row == null ) {
				IConfiguration c = crow.addChild( "fo:table-cell" );
				c.addChild( "fo:block" );
				continue;
			}
			double h = (double )row.getHeight() / 700.0;
			crow.setAttribute( "height", String.valueOf( h ) + "cm" );
			for ( int j = 0 ; j < nMaxCol; j++ ) {
				if ( pos[i][j] < 0 ) {
					continue;
				}
				IConfiguration cell = crow.addChild( "fo:table-cell" );
				HSSFCell hcell = row.getCell( j );
				if ( hcell == null ) {
					cell.addChild( "fo:block" );
					continue;
				}
				boolean bb = false;
				if ( i > 0 ) {
					if ( bborder[i-1][j] != 0 ) {
						bb = true;
					} else {
						HSSFRow prow = sheet.getRow( i-1 );
						if ( prow != null ) {
							HSSFCell pcell = prow.getCell( j );
							if ( pcell != null ) {
								HSSFCellStyle pst = pcell.getCellStyle();
								if ( pst.getBorderBottom() != HSSFCellStyle.BORDER_NONE ) {
									bb = true;
								}
							}
						}
					}
				}
				boolean rb = false;
				if ( j > 0 ) {
					if ( bborder[i][j-1] != 0 ) {
						rb = true;
					} else {
						HSSFCell pcell = row.getCell( j-1 );
						if ( pcell != null ) {
							HSSFCellStyle pst = pcell.getCellStyle();
							if ( pst.getBorderRight() != HSSFCellStyle.BORDER_NONE ) {
								rb = true;
							}
						}
					}
				}
				setupCell( cell, book, hcell, format, rb, bb );
				if ( pos[i][j] > 0 ) {
					CellRangeAddress reg = findRegion( sheet, i, j );
					if ( reg.getFirstColumn() != reg.getLastColumn()) {
						cell.setAttribute( "number-columns-spanned",
							String.valueOf( reg.getLastColumn() - reg.getFirstColumn() + 1 ));
					}
					if ( reg.getFirstRow() != reg.getLastRow()) {
						cell.setAttribute( "number-rows-spanned",
						String.valueOf( reg.getLastRow() - reg.getFirstRow() + 1 ));
					}
				}   
			}
		}

	}

	private IConfiguration addPageMaster( IConfiguration master,
				String name, HSSFPrintSetup setup )
	{
		String	size = "";
		String	height = "", width = "";
		switch ( setup.getPaperSize()) {
		case HSSFPrintSetup.A4_PAPERSIZE:
			size = "A4";
			width = "29.7cm";
			height = "21cm";
			break;
		case HSSFPrintSetup.A5_PAPERSIZE:
			size = "A5";
			width = "21cm";
			height = "14.8cm";
			break;
		case HSSFPrintSetup.ENVELOPE_10_PAPERSIZE:
			size = "ENVELOPE_10";
			break;
		case HSSFPrintSetup.ENVELOPE_CS_PAPERSIZE:
			size = "ENVELOPE_CS";
			break;
		case HSSFPrintSetup.ENVELOPE_DL_PAPERSIZE:
			size = "ENVELOPE_DL";
			break;
		case HSSFPrintSetup.ENVELOPE_MONARCH_PAPERSIZE:
			size = "ENVELOPE_MONARCH";
			break;
		case HSSFPrintSetup.EXECUTIVE_PAPERSIZE:
			size = "EXECUTIVE";
			break;
		case HSSFPrintSetup.LEGAL_PAPERSIZE:
			size = "LEGAL";
			width = "35.56cm";
			height = "21.59cm";
			break;
		case HSSFPrintSetup.LETTER_PAPERSIZE:
			size = "LETTER";
			width = "27.94cm";
			height = "21.59cm";
			break;
		case 8: // A3
			size = "A3";
			width = "42cm";
			height = "29.7cm";
			break;
		case 12: // B4(JIS)
			size = "B4";
			width = "36.4cm";
			height = "25.7cm";
			break;
		case 13: // B5(JIS)
			size = "B5";
			width = "25.7cm";
			height = "18.2cm";
			break;
		case 33: // B4(ISO)
			size = "B4";
			width = "35.3cm";
			height = "25cm";
			break;
		case 34: // B5(ISO)
			size = "B5";
			width = "25cm";
			height = "17.6cm";
			break;
		default:
			width = "29.7cm";
			height = "21cm";
			size = "unknwon size(" + setup.getPaperSize() + ")";
		}
		scale = setup.getScale();

		if ( isDebug()) {
			logger.debug( "addPage[" + name + "]"
				+ " size=" + size + " height=" + height + " width=" + width
				+ " orientation=" + ( setup.getLandscape() ? "LANDSCAPE" : "PORTRAIT" )
				+ " height=" + setup.getFitHeight()
				+ " width=" + setup.getFitWidth()
				+ " option=" + setup.getOptions()
				+ " scale=" + setup.getScale()
				+ " hres=" + setup.getHResolution()
				+ " vres=" + setup.getVResolution() );
		}
		IConfiguration layout = master.addChild( "fo:simple-page-master" );
		layout.setAttribute( "master-name", name );
		if ( setup.getLandscape()) {
			layout.setAttribute( "page-height", height );
			layout.setAttribute( "page-width", width );
		} else {
			layout.setAttribute( "page-height", width );
			layout.setAttribute( "page-width", height );
		}
		layout.setAttribute( "margin-top", "1cm" );
		layout.setAttribute( "margin-bottom", "1cm" );
		layout.setAttribute( "margin-right", "1cm" );
		layout.setAttribute( "margin-left", "1cm" );
		IConfiguration c;
		c = layout.addChild( "fo:region-body" );
		return layout;
	}

	private CellRangeAddress findRegion( HSSFSheet sheet, int row, int col ) {
		for ( int i = 0; i < sheet.getNumMergedRegions(); i++ ) {
			CellRangeAddress reg = sheet.getMergedRegion( i );
			if ( row == reg.getFirstRow() && col == reg.getFirstColumn()) {
				return reg;
			}
		}
		return null;
	}

	private boolean isRowBreak( HSSFSheet sheet, int row ) {
		int[] rb = sheet.getRowBreaks();
		if ( rb == null ) {
			return false;
		}
		for ( int i = 0; i < rb.length; i++ ) {
			if ( row == rb[i] ) {
				return true;
			}
		}
		return false;
	}

	private void setupCell( IConfiguration cell,
				HSSFWorkbook book, HSSFCell hcell, HSSFDataFormat format,
				boolean rborder, boolean bborder )
	{
		HSSFCellStyle hstyle = hcell.getCellStyle();
		if ( !rborder ) {
			setBorder( cell, "border-left",
				hstyle.getBorderLeft(), hstyle.getLeftBorderColor() );
		}
		if ( !bborder ) {
			setBorder( cell, "border-top",
				hstyle.getBorderTop(), hstyle.getTopBorderColor() );
		}
		setBorder( cell, "border-right",
			hstyle.getBorderRight(), hstyle.getRightBorderColor() );
		setBorder( cell, "border-bottom",
			hstyle.getBorderBottom(), hstyle.getBottomBorderColor() );
		String bgcol = convertColor( hstyle.getFillForegroundColor(),
									hstyle.getFillBackgroundColor() );
		if ( bgcol != null && !bgcol.equals( "AUTO" )) {
			cell.setAttribute( "background-color", bgcol );
		}

		String	align = null;
		switch ( hstyle.getAlignment() ) {
		case HSSFCellStyle.ALIGN_RIGHT: align = "right"; break;
		case HSSFCellStyle.ALIGN_CENTER: align = "center"; break;
		case HSSFCellStyle.ALIGN_JUSTIFY: align = "justify"; break;
		case HSSFCellStyle.ALIGN_CENTER_SELECTION: align = "center"; break;
		case HSSFCellStyle.ALIGN_LEFT:
		default:
		}

		String	valign = null;
		switch ( hstyle.getVerticalAlignment() ) {
		case HSSFCellStyle.VERTICAL_TOP: break;
		case HSSFCellStyle.VERTICAL_BOTTOM: valign = "after"; break;
		default:
		case HSSFCellStyle.VERTICAL_JUSTIFY:
		case HSSFCellStyle.VERTICAL_CENTER: valign = "center"; break;
		}

		HSSFFont hfont = book.getFontAt( hcell.getCellStyle().getFontIndex());
		int fsize = hfont.getFontHeight() / 24;
//		int fsize = hfont.getFontHeightInPoints();
		String str = getString( hcell, format );
		if ( str != null && str.length() > 0 ) {
			IConfiguration blk = null;
			if ( hstyle.getRotation() == -165 ) {
				IConfiguration blkc = cell.addChild( "fo:block-container" );
				blkc.setAttribute( "writing-mode", "tb-rl" );
				blk = blkc.addChild( "fo:block" );
			} else {
				blk = cell.addChild( "fo:block" );
			}
			blk.set( str );
			blk.setAttribute( "font-family", fontfamily );
			blk.setAttribute( "font-size", fsize + "pt" );
			if ( hfont.getBoldweight() == HSSFFont.BOLDWEIGHT_NORMAL ) {
				blk.setAttribute( "font-weight", "normal" );
			} else {
				blk.setAttribute( "font-weight", "bold" );
			}
			if ( hfont.getItalic()) {
				blk.setAttribute( "font-style", "italic" );
			}
			String deco = "";
			if ( hfont.getStrikeout() ) {
				deco = "line-through";
			}
			if ( hfont.getUnderline() != HSSFFont.U_NONE) {
				deco = "underline";
			}
			if ( deco != null && deco.length() > 0 ) {
				blk.setAttribute( "text-decoration", deco );
			}
			String fgcol = convertColor( hfont.getColor() );
			if ( fgcol != null && !fgcol.equals( "AUTO" )) {
				blk.setAttribute( "color", fgcol );
			}
			if ( align != null ) {
				blk.setAttribute( "text-align", align );
			}
			if ( valign != null ) {
				cell.setAttribute( "display-align", valign );
			}
		} else {
			IConfiguration blk = cell.addChild( "fo:block" );
		}
	}

	private void setBorder( IConfiguration c, String tag, short sType, short sColor ) {
		String style = null, color, width;
		color = "black";
		width = "thin";
		switch ( sType ) {
		case HSSFCellStyle.BORDER_NONE:
			return;
		case HSSFCellStyle.BORDER_DASH_DOT:
		case HSSFCellStyle.BORDER_DASH_DOT_DOT:
		case HSSFCellStyle.BORDER_DASHED:
			style = "dashed";
			width = "1pt";
			break;
		case HSSFCellStyle.BORDER_DOTTED:
			style = "dotted";
			width = "1pt";
			break;
		case HSSFCellStyle.BORDER_DOUBLE:
			style = "double";
			width = "1pt";
			break;
		case HSSFCellStyle.BORDER_HAIR:
			style = "solid";
			width = "0.5pt";
			break;
		case HSSFCellStyle.BORDER_MEDIUM:
			style = "solid";
			width = "2pt";
			break;
		case HSSFCellStyle.BORDER_MEDIUM_DASH_DOT:
		case HSSFCellStyle.BORDER_MEDIUM_DASH_DOT_DOT:
		case HSSFCellStyle.BORDER_MEDIUM_DASHED:
		case HSSFCellStyle.BORDER_SLANTED_DASH_DOT:
			style = "dashed";
			width = "2pt";
			break;
		case HSSFCellStyle.BORDER_THICK:
			style = "solid";
			width = "3pt";
			break;
		case HSSFCellStyle.BORDER_THIN :
			style = "solid";
			width = "0.5pt";
			break;
		}
		c.setAttribute( tag, width + " " + style + " " + color );
	}

	private String convertColor( short fg, short bg ) {
		String ret = convertColor( fg );
		if ( ret == null ) {
			ret = convertColor( bg );
		}
		return ret;
	}

	private String convertColor( short idx ) {
		String ret = null;
		switch ( idx ) {
		case HSSFColor.AUTOMATIC.index: ret = "AUTO"; break;
		case HSSFColor.BLACK.index: ret = "black"; break;
		case HSSFColor.BLUE.index: ret = "blue"; break;
		case HSSFColor.GREEN.index: ret = "green"; break;
		case HSSFColor.ORANGE.index: ret = "orange"; break;
		case HSSFColor.PINK.index: ret = "pink"; break;
		case HSSFColor.RED.index: ret = "red"; break;
		case HSSFColor.WHITE.index: ret = "white"; break;
		case HSSFColor.YELLOW.index: ret = "yellow"; break;
		}
		if ( ret != null ) {
			return ret;
		}
		Hashtable hash = HSSFColor.getIndexHash();
		HSSFColor col = (HSSFColor )hash.get( Integer.valueOf( idx ));
		if ( col == null ) {
			return null;
		}
		short[] trip = col.getTriplet();
		ret = "rgb(" + trip[0] + "," + trip[1] + "," + trip[2] + ")";
		return ret;
	}

	private String getString( HSSFCell hcell, HSSFDataFormat format ) {
		DecimalFormat df1 = null;
		WarekiFormat df2 = null;
		String str = null;
		switch ( hcell.getCellType()) {
		case HSSFCell.CELL_TYPE_STRING:
			str = hcell.getRichStringCellValue().getString();
			break;
		case HSSFCell.CELL_TYPE_NUMERIC:
			short idx = hcell.getCellStyle().getDataFormat();
			double val = hcell.getNumericCellValue();
			switch ( hcell.getCellStyle().getDataFormat()) {
			case 0x05:
				df1 = new DecimalFormat( "\\#,##0" );
			case 0x0:
			case 0xa4:
				if ( df1 == null ) {
					df1 = new DecimalFormat( "#" );
				}
				str = df1.format( val );
				break;
			case 0xb0:
				if ( df2 == null ) {
					df2 = new WarekiFormat( "gggY年M月d日" );
				}
			case 0xa5:
				if ( df2 == null ) {
					df2 = new WarekiFormat( "yyyy/M/d" );
				}
			case 0x0e:
				if ( df2 == null ) {
					df2 = new WarekiFormat( "M/d/yy" );
				}
				Date date = new Date( (long )(val - dayDiff) * 60 * 60 * 24 * 1000 );
				str = df2.format( date );
				break;
			default:
				str = String.valueOf( hcell.getNumericCellValue() );
			}
			break;
		case HSSFCell.CELL_TYPE_FORMULA:
			str = hcell.getCellFormula();
			break;
		}
		return str;
	}

	/**
	* サーブレットのContentTypeヘッダを取得します。
	* @return	サーブレットのContentTypeヘッダ
	*/
	public String getContentType() {
		return "application/pdf";
	}

	/**
	* ファイルの拡張子を返します。
	* @return   ファイルの拡張子
	*/
	public String getExtension() {
		return "pdf";
	}

}
