package	jp.co.sds_corp.report;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.net.URL;
import java.sql.Timestamp;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import jp.co.sds_corp.report.source.SourceFactory;
import jp.co.sds_corp.util.common.BeanParser;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;
import jp.co.sds_corp.util.script.ScriptParserFactory;

import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFRichTextString;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.CellRangeAddress;
import org.apache.poi.hssf.util.CellReference;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;

/**
<a href="http://jakarta.apache.org/poi/">Jakarta POI</a>を使用して、Excel帳票を出力します。
<pre>
&lt;Report
	name="ReportTest"
	type="Excel"
	template="${Report.dir}/table.xls"
	sheet="Sheet1"&gt;
	&lt;Script ... /&gt;
	&lt;Source ... /&gt;
&lt;/Report&gt;
</pre>
Source要素については、{@link SourceFactory}を参照してください。
Script要素については、{@link ScriptParserFactory}を参照してください。
*/
public class ExcelReport extends AbstractReport
{
	CmnLogger logger = CmnLogFactory.getInstance().getLogger(this);
	String	template;
	String	sheetName;
	String	orgSheetName;
	String	script;
	boolean	protect = false;
	String	protectPwd = null;
	HashMap<String,HSSFSheet>	mapTemp = new HashMap<String,HSSFSheet>();

	protected HSSFWorkbook result = null;
	protected HSSFSheet sheet = null;
	protected HSSFWorkbook tmpBook = null;
	protected HSSFSheet orgSheet = null;
	boolean bPrint = false;

	/**
	* 設定情報で初期化します。
	* @param	config	設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		template = config.getAttributes().get( "template", null );
		if ( getProperty() != null ) {
			template = getProperty().replaceValue( template );
		}
		sheetName = config.getAttributes().get( "sheet", null );
		protectPwd = config.getAttributes().get( "protect", null );
		if ( protectPwd != null ) {
			protect = true;
		}

		if ( template != null ) {
			loadFile( template );
			if ( sheetName != null ) {
				loadPage( sheetName );
			}
		}
	}

	/**
	* 処理を終了します。
	*/
	public void exit() throws Exception {
		result = null;
		sheet = null;
		tmpBook = null;
		orgSheet = null;
	}

	/**
	* 帳票の出力結果をオブジェクトとして取得します。
	* @return	出力結果
	*/
	public Object getResult() {
		return result;
	}

	/**
	* 帳票を出力します。
	*/
	public void print() throws Exception {
		super.print();
		Iterator<String> itr = mapTemp.keySet().iterator();
		while ( itr.hasNext()){
			String name = (String )itr.next();
			int idx = result.getSheetIndex( name );
			if ( isDebug() ) {
				logger.debug( "remove template " + name + " (" + idx );
			}
			if ( idx >= 0 ) {
				result.removeSheetAt( result.getSheetIndex( name ));
			}
		}
		if ( protect ) {
			for ( int i = 0; i < result.getNumberOfSheets(); i++ ) {
				HSSFSheet sheet = result.getSheetAt( i );
				sheet.protectSheet( protectPwd );
			}
		}
		bPrint = true;
	}

	/**
	* 帳票を出力します。
	* @param	out		出力先
	*/
	public void print( OutputStream out ) throws Exception {
		if ( !bPrint ) {
			print();
		}
		result.write( out );
	}

	/**
	* サーブレットのContentTypeヘッダを取得します。
	* @return	サーブレットのContentTypeヘッダ
	*/
	public String getContentType() {
		return "application/vnd.ms-excel";
	}

	/**
	* ファイルの拡張子を返します。
	* @return   ファイルの拡張子
	*/
	public String getExtension() {
		return "xls";
	}

	/**
	* templateシートの名前を取得します。
	* @param	name	シート名
	*/
	protected String getTemplateName( String name ) {
//		return "__" + name + "__";
		return name;
	}

	/**
	* ファイルを読み込みます。
	* @param	template	ファイル名
	*/
	public void loadFile( String template ) throws Exception {
		if ( isDebug()) {
			logger.debug( "load file " + template );
		}
		InputStream is = null;
		try {
			is = new FileInputStream( new File( template ));
		} catch ( Exception ex ) {
			URL url = new URL( template );
			is = url.openStream();
		}
		tmpBook = createHSSFWorkbook( is );
		is.close();
		result = tmpBook;
		mapTemp.clear();
		for ( int i = 0; i < tmpBook.getNumberOfSheets(); i++ ) {
			if ( tmpBook.getSheetName( i ).equals( "000000" )) {
				if ( isDebug() ) {
					logger.debug( "ignore template " + i + ") "
							+ tmpBook.getSheetName( i ));
				}
				continue;
			}
			String name = getTemplateName( tmpBook.getSheetName( i ) );
			if ( isDebug() ) {
				logger.debug( "load template " + i + ") "
							+ tmpBook.getSheetName( i ) + "->" + name );
			}
/*
// @@@ POI-3.0用
			result.setSheetName( i, name );
// @@@ POI-2.5.1用
//			result.setSheetName( i, name, HSSFWorkbook.ENCODING_UTF_16 );
*/
			mapTemp.put( name, tmpBook.getSheetAt( i ));
		}
	}

	private HSSFWorkbook createHSSFWorkbook( InputStream is ) throws Exception {
		POIFSFileSystem filein = new POIFSFileSystem( is );
		return new HSSFWorkbook( filein );
	}

	/**
	* ページを読み込みます。
	* @param	sheetName	ページ名
	*/
	public void loadPage( String sheetName ) throws Exception {
		if ( isDebug() ) {
			logger.debug( "load page " + sheetName );
		}
		this.sheetName = sheetName;
		this.orgSheetName = sheetName;
		sheet = result.getSheet( sheetName );
		if ( sheet == null ) {
			throw( new Exception( "Sheet[" + sheetName + "] not found." ));
		}
		orgSheet = tmpBook.getSheet( orgSheetName );
		if ( orgSheet == null ) {
			throw( new Exception( "Sheet[" + orgSheetName + "] not found." ));
		}
		mapTemp.remove( sheetName );
	}
	/**
	* ページに名前をつけて読み込みます。
	* @param	newName	新しいページ名
	* @param	orgName	読み込むページ名
	*/
	public void loadPage( String newName,String orgName ) throws Exception {
		orgSheetName = getTemplateName( orgName );
		if ( isDebug() ) {
			logger.debug( "load page " + orgSheetName + "->" + newName );
		}
		sheet = null;
		orgSheet = tmpBook.getSheet( orgSheetName );
		if ( orgSheet == null ) {
			throw( new Exception( "Sheet[" + orgSheetName + "] not found." ));
		}
		int idx = result.getSheetIndex( orgSheetName );
		if ( idx >= 0 ) {
			sheet = result.cloneSheet( idx );
			copyColumnWidth( sheet, result.getSheetAt( idx ));
			int nLoad = result.getNumberOfSheets() - 1;
// @@@ POI-3.0用
			result.setSheetName( nLoad, newName );
// @@@ POI-2.5.1用
//			result.setSheetName( nLoad, newName, HSSFWorkbook.ENCODING_UTF_16 );
			sheetName = newName;
		} else {
			throw( new IllegalArgumentException( "Page [" + orgName + "] not found. " + idx ));
		}
	}

	private void copyColumnWidth( HSSFSheet dst, HSSFSheet src ) {
		for ( int i = 0; i < 256; i++ ) {
			int width = src.getColumnWidth( i );
			if ( width == 8 ) {
				width = 2304;
			}
			dst.setColumnWidth( i, width );
		}
	}

	/**
	* 現在処理中のページを、指定されたページに追加します。
	* @param	dstName	追加するページ
	*/
	public void appendTo( String dstName ) throws Exception {
		appendTo( dstName, true );
	}
	/**
	* 現在処理中のページを、指定されたページに追加します。
	* @param	dstName	追加するページ
	* @param	bBreak	改ページフラグ
	*/
	public void appendTo( String dstName, boolean bBreak ) throws Exception {
		if ( sheetName.equals( dstName )) {
			throw( new Exception( "cannot append to same sheet["
									+ dstName + "]" ));
		}
		HSSFSheet dst = result.getSheet( dstName );
		if ( dst == null ) {
			if ( isDebug() ) {
				logger.debug( "rename " + sheetName + "->" + dstName );
			}
			int idx = result.getSheetIndex( sheetName );
// @@@ POI-3.0用
			result.setSheetName( idx, dstName );
// @@@ POI-2.5.1用
//			result.setSheetName( idx, dstName, HSSFWorkbook.ENCODING_UTF_16 );
			return;
		}
		int nTo = dst.getLastRowNum() + 1;
		int nRow = sheet.getLastRowNum() + 1;
		if ( isDebug() ) {
			logger.debug( "appendTo " + sheetName + " " + nRow
							+ "->" + dstName + " " + nTo );
		}
		for ( int i = 0; i < nRow; i++ ) {
			HSSFRow row = sheet.getRow( i );
			HSSFRow newRow = dst.createRow( nTo + i );
			copyRow( newRow, row );
		}
		for ( int i = 0; i < sheet.getNumMergedRegions(); i++ ) {
			CellRangeAddress reg = sheet.getMergedRegion( i );
			reg.setFirstRow( reg.getFirstRow() + nTo );
			reg.setLastRow( reg.getLastRow() + nTo );
			dst.addMergedRegion( reg );
		}
		if ( bBreak ) {
			dst.setRowBreak( dst.getLastRowNum() );
		}
		int idx = result.getSheetIndex( sheetName );
		result.removeSheetAt( idx );
	}

	/**
	* 現在処理中のページに改ページを追加します。
	*/
	public void endPage() {
		endPage( sheet.getLastRowNum() );
	}
	/**
	* 現在処理中のページに改ページを追加します。
	* @param	row		改ページを追加する行
	*/
	public void endPage( int row ) {
		if ( isDebug()) {
			logger.debug( "end page " + row );
		}
		sheet.setRowBreak( row );
	}

	/**
	* ページの内容をコピーします。
	* @param	dst		コピー先
	* @param	src		コピー元
	*/
	protected void copySheet( HSSFSheet dst, HSSFSheet src ) {
		for ( int i = 0; i <= src.getLastRowNum(); i++ ) {
			HSSFRow srcRow = src.getRow( i );
			HSSFRow dstRow = dst.createRow( i );
			copyRow( dstRow, srcRow );
		}
		for ( int i = 0; i < 255; i++ ) {
			dst.setColumnWidth( i, src.getColumnWidth( i ));
		}
		for ( int i = 0; i < src.getNumMergedRegions(); i++ ) {
			CellRangeAddress reg = src.getMergedRegion( i );
			dst.addMergedRegion( reg );
		}
		BeanParser parser;
		parser = new BeanParser( src );
		parser.copyTo( dst );
		parser = new BeanParser( src.getPrintSetup() );
		parser.copyTo( dst.getPrintSetup() );
	}

	/**
	* 行の内容をコピーします。
	* @param	dst		コピー先
	* @param	src		コピー元
	*/
	protected void copyRow( HSSFRow dst, HSSFRow src ) {
		if ( src == null ) {
			return;
		}
		dst.setHeight( src.getHeight());
		dst.setHeightInPoints( src.getHeightInPoints());
		for ( int i = 0; i < src.getLastCellNum(); i++ ) {
			HSSFCell srcCell = src.getCell( i );
			HSSFCell dstCell = dst.createCell( i );
			copyCell( dstCell, srcCell );
		}
	}

	/**
	* セルの内容をコピーします。
	* @param	dst		コピー先
	* @param	src		コピー元
	*/
	protected void copyCell( HSSFCell dst, HSSFCell src ) {
		if ( src == null ) {
			return;
		}
		switch ( src.getCellType()) {
		case HSSFCell.CELL_TYPE_BLANK :
			break;
		case HSSFCell.CELL_TYPE_BOOLEAN :
			dst.setCellValue( src.getBooleanCellValue());
			break;
		case HSSFCell.CELL_TYPE_ERROR :
			dst.setCellValue( src.getErrorCellValue());
			break;
		case HSSFCell.CELL_TYPE_FORMULA :
			dst.setCellFormula( src.getCellFormula());
			break;
		case HSSFCell.CELL_TYPE_NUMERIC :
			dst.setCellValue( src.getNumericCellValue());
			break;
		case HSSFCell.CELL_TYPE_STRING :
// @@@ POI-3.0用
			dst.setCellValue( src.getRichStringCellValue());
// @@@ POI-2.5.1用
//			dst.setCellValue( src.getStringCellValue());
			break;
		default :
		}
		// スタイルを取得
		// dst.setCellStyle( src.getCellStyle());
		HSSFCellStyle style = src.getCellStyle();
/*
		HSSFCellStyle newStyle = result.createCellStyle();
		BeanParser parser = new BeanParser( style );
		parser.copyTo( newStyle );
		newStyle.setFont( tmpBook.getFontAt( style.getFontIndex()));
		dst.setCellStyle( newStyle );
*/
		dst.setCellStyle( style );
	}

	/**
	* セルから値を取得します。
	* @param	pos		セルの位置
	* @return	設定されている値
	*/
	public Object getCell( String pos ) throws Exception {
		CellReference ref = new CellReference( pos );
		return getCell( ref.getCol(), ref.getRow() );
	}
	/**
	* セルから値を取得します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @return	設定されている値
	*/
	public Object getCell( int col, int row ) throws Exception {
		HSSFRow hrow = sheet.getRow( row );
		if ( hrow == null ) {
			throw( new Exception() );
		}
		HSSFCell cell = hrow.getCell( col );
		if ( cell == null ) {
			throw( new Exception() );
		}
		Object ret = null;
		switch ( cell.getCellType()) {
		case HSSFCell.CELL_TYPE_BLANK :
			break;
		case HSSFCell.CELL_TYPE_BOOLEAN :
			ret = Boolean.valueOf( cell.getBooleanCellValue() );
			break;
		case HSSFCell.CELL_TYPE_ERROR :
			ret = Byte.valueOf( cell.getErrorCellValue() );
			break;
		case HSSFCell.CELL_TYPE_FORMULA :
			ret = cell.getCellFormula();
			break;
		case HSSFCell.CELL_TYPE_NUMERIC :
			ret = Double.valueOf( cell.getNumericCellValue() );
			break;
		case HSSFCell.CELL_TYPE_STRING :
// @@@ POI-3.0用
			ret = cell.getRichStringCellValue();
// @@@ POI-2.5.1用
//			ret = cell.getStringCellValue();
			break;
		default :
		}
		if ( isDebug()) {
			String pos = new CellReference( row, col ).toString();
			logger.debug( "getCell " + sheetName + "."
								+ pos + "(" + col + "," + row + ")=" + ret );
		}
		return ret;
	}

	/**
	* セルに値を設定します。
	* @param	pos		セルの位置
	* @param	data	設定する値
	*/
	public void setCell( String pos, Object data ) throws Exception {
		CellReference ref = new CellReference( pos );
		setCell( ref.getCol(), ref.getRow(), data );
	}

	/**
	* セルに値を設定します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	data	設定する値
	*/
	public void setCell( int col, int row, Object data ) throws Exception {
		boolean bRow = false;
		HSSFRow hrow = sheet.getRow( row );
		if ( hrow == null ) {
			bRow = true;
			hrow = sheet.createRow( row );
		}
		boolean bCell = false;
		HSSFCell cell = hrow.getCell( col );
		if ( cell == null ) {
			bCell = true;
			cell = hrow.createCell( col );
		}
		if ( isDebug() ) {
			String pos = new CellReference( row, col ).toString();
			String name = ( data == null ? "null" : data.getClass().getName());
			String msg = "setCell " + sheetName + "."
				+ pos + "(" + col + "," + row + ")=" + data
				+ " ;" + name;
			if ( bRow ) {
				msg += " R";
			}
			if ( bCell ) {
				msg += " C";
			}
			logger.debug( msg );
		}

		if ( data == null ) {
			// return;
			data = "";
		}
// @@@ POI-2.5.1用
//		cell.setEncoding( HSSFCell.ENCODING_UTF_16 );
		if ( data instanceof java.lang.String ) {
// @@@ POI-3.0用
			cell.setCellValue( new HSSFRichTextString( data.toString()));
// @@@ POI-2.5.1用
//			cell.setCellValue( data.toString());
		} else 
		if ( data instanceof java.lang.Boolean ) {
		} else 
		if ( data instanceof java.lang.Integer ) {
			cell.setCellValue( ((Integer )data ).intValue());
		} else 
		if ( data instanceof java.lang.Long ) {
		} else 
		if ( data instanceof java.lang.Float ) {
		} else 
		if ( data instanceof java.lang.Double ) {
			cell.setCellValue( ((Double )data).doubleValue());
		} else
		if ( data instanceof java.math.BigDecimal ) {
			cell.setCellValue( ((BigDecimal )data).doubleValue());
		} else
		if ( data instanceof java.math.BigInteger ) {
		} else
		if ( data instanceof java.sql.Timestamp ) {
			/*
			cell.setCellValue( new Date(((java.sql.Timestamp )data).getTime()));
			*/
			cell.setCellValue( (Timestamp )data );
		} else {
			throw new IllegalArgumentException( "データ型が不正です:" + data.getClass().getName());
		}
	}

	/**
	* セルに式を設定します。
	* @param	pos		セルの位置
	* @param	formula	設定する式
	*/
	public void setCellFormula( String pos, String formula ) throws Exception {
		CellReference ref = new CellReference( pos );
		setCellFormula( sheet, ref.getCol(), ref.getRow(), formula );
	}
	/**
	* セルに式を設定します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	formula	設定する式
	*/
	public void setCellFormula( int col, int row, String formula )
		throws Exception
	{
		setCellFormula( sheet, col, row, formula );
	}

	/**
	* セルに式を設定します。
	* @param	sheet	ページ
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	formula	設定する式
	*/
	public void setCellFormula( HSSFSheet sheet, int col, int row, String formula )
		throws Exception
	{
		boolean bRow = false;
		HSSFRow hrow = sheet.getRow( row );
		if ( hrow == null ) {
			bRow = true;
			hrow = sheet.createRow( row );
		}
		boolean bCell = false;
		HSSFCell cell = sheet.getRow( row ).getCell( col );
		if ( cell == null ) {
			bCell = true;
			cell = hrow.createCell( col );
		}
		if ( isDebug() ) {
			String pos = new CellReference( row, col ).toString();
			String msg = "formula " + sheetName + "."
				+ pos + "(" + col + "," + row + ")=" + formula;
			if ( bRow ) {
				msg += " R";
			}
			if ( bCell ) {
				msg += " C";
			}
			logger.debug( msg );
		}
		cell.setCellFormula( formula );
	}

	/**
	* セルに罫線を設定します。
	* @param	pos		セルの位置
	* @param	location	罫線の位置
	* @param	border	設定する罫線
	*/
	public void setBorder( String pos, int location, Border border )
		throws Exception
	{
		CellReference ref = new CellReference( pos );
		setBorder( ref.getCol(), ref.getRow(), location, border );
	}
	/**
	* セルに式を設定します。
	* @param	col		セルの桁
	* @param	row		セルの行
	* @param	location	罫線の位置
	* @param	border	設定する罫線
	*/
	public void setBorder( int col, int row, int location, Border border )
		throws Exception
	{
		if ( isDebug()) {
			String pos = new CellReference( row, col ).toString();
			logger.debug( "border " + sheetName + "."
				+ pos + "(" + col + "," + row + ")." + location + "=" + border );
		}
		HSSFRow dstRow = sheet.getRow( row );
		if ( dstRow == null ) {
			dstRow = sheet.createRow( row );
		}
		HSSFCell dstCell = dstRow.getCell( col );
		if ( dstCell == null ) {
			dstCell = dstRow.createCell( col );
		}
		HSSFCellStyle orgStyle = dstCell.getCellStyle();
		HSSFCellStyle newStyle = result.createCellStyle();
		BeanParser parser;
		parser = new BeanParser( orgStyle );
		parser.copyTo( newStyle );
		newStyle.setFont( result.getFontAt( orgStyle.getFontIndex() ));
		short pb = convertBorder( border );
		switch ( location ) {
		case Border.TOP:
			newStyle.setBorderTop( pb );
			break;
		case Border.BOTTOM:
			newStyle.setBorderBottom( pb );
			break;
		case Border.LEFT:
			newStyle.setBorderLeft( pb );
			break;
		case Border.RIGHT:
			newStyle.setBorderRight( pb );
			break;
		}
		dstCell.setCellStyle( newStyle );
	}

	protected short convertBorder( Border border ) {
		short ret;
		switch ( border.getType() ){
		case Border.TYPE_LINE:
			ret = HSSFCellStyle.BORDER_THIN;
			break;
		case Border.TYPE_DOT:
			ret = HSSFCellStyle.BORDER_DOTTED;
			break;
		case Border.TYPE_DASH_DOT:
			ret = HSSFCellStyle.BORDER_DASH_DOT;
			break;
		case Border.TYPE_DOUBLE:
			ret = HSSFCellStyle.BORDER_DOUBLE;
			break;
		case Border.TYPE_NONE:
		default:
			ret = HSSFCellStyle.BORDER_NONE;
		}
		return ret;
	}

	/**
	* セルの値をコピーします。
	* @param	fpos	コピー元のセルの位置
	* @param	tpos	コピー先のセルの位置
	*/
	public void copyCell( String fpos, String tpos ) throws Exception {
		CellReference fref = new CellReference( fpos );
		CellReference tref = new CellReference( tpos );
		copyCell( fref.getCol(), fref.getRow(), tref.getCol(), tref.getRow() );
	}
	/**
	* セルの値をコピーします。
	* @param	fcol	コピー元のセルの桁
	* @param	frow	コピー元のセルの行
	* @param	tcol	コピー先のセルの桁
	* @param	trow	コピー先のセルの行
	*/
	public void copyCell( int fcol, int frow, int tcol, int trow )
		throws Exception
	{
		if ( isDebug()) {
			String fpos = new CellReference( frow, fcol ).toString();
			String tpos = new CellReference( trow, tcol ).toString();
			logger.debug( "copy "
				+ orgSheetName + "." + fpos + "(" + fcol + "," + frow + ")->"
				+ sheetName + "." + tpos + "(" + tcol + "," + trow + ")" );
		}
		HSSFRow srcRow = orgSheet.getRow( frow );
		HSSFRow dstRow = sheet.getRow( trow );
		if ( dstRow == null ) {
			dstRow = sheet.createRow( trow );
		}
		HSSFCell srcCell = srcRow.getCell( fcol );
		HSSFCell dstCell = dstRow.getCell( tcol );
		if ( dstCell == null ) {
			dstCell = dstRow.createCell( tcol );
		}
		copyCell( dstCell, srcCell );

		// スタイルを取得
		HSSFCellStyle style = srcCell.getCellStyle();
		dstCell.setCellStyle( style );
	}

	/**
	* セルをクリアします。
	* @param	pos		セルの位置
	*/
	public void clearCell( String pos ) throws Exception {
		CellReference ref = new CellReference( pos );
		clearCell( ref.getCol(), ref.getRow() );
	}
	/**
	* セルをクリアします。
	* @param	col		コピー元のセルの桁
	* @param	row		コピー元のセルの行
	*/
	public void clearCell( int col, int row ) throws Exception {
		if ( isDebug()) {
			String pos = new CellReference( row, col ).toString();
			logger.debug( "clear " + sheetName + "."
				+ pos + "(" + col + "," + row + ")" );
		}
		setCell( col, row, "" );
		HSSFRow dstRow = sheet.getRow( row );
		HSSFCell dstCell = dstRow.getCell( col );
		HSSFCellStyle newStyle = result.createCellStyle();
		dstCell.setCellStyle( newStyle );
	}

	/**
	* セルを結合します。
	* @param	fpos	開始位置
	* @param	tpos	終了位置
	*/
	public void mergeCell( String fpos, String tpos )
		throws Exception
	{
		CellReference fref = new CellReference( fpos );
		CellReference tref = new CellReference( tpos );
		mergeCell( fref.getCol(), fref.getRow(), tref.getCol(), tref.getRow() );
	}
	/**
	* セルを結合します。
	* @param	colFr	開始位置
	* @param	rowFr	開始位置
	* @param	colTo	終了位置
	* @param	rowTo	終了位置
	*/
	public void mergeCell( int colFr, int rowFr, int colTo, int rowTo )
		throws Exception
	{
		if ( isDebug()) {
			String posFr = new CellReference( rowFr, colFr ).toString();
			String posTo = new CellReference( rowTo, colTo ).toString();
			logger.debug( "merge cell " + posFr + "," + posTo );
		}
		CellRangeAddress reg = new CellRangeAddress( rowFr, rowTo, colFr, colTo );
		sheet.addMergedRegion( reg );
	}

	/**
	* 現在のページに行を追加します。
	* @param	row		追加する位置
	*/
	public void addRow( int row ) throws Exception {
		if ( isDebug()) {
			logger.debug( "add row " + row );
		}
		sheet.createRow( row );
	}
	/**
	* 現在のページに、templateから書式をコピーして行を追加します。
	* @param	dst		追加する位置
	* @param	src		書式の行の位置
	*/
	public void addRow( int dst, int src ) throws Exception {
		HSSFRow srcRow = orgSheet.getRow( src );
		addRow( dst, src, 0, srcRow.getLastCellNum() );
	}
	/**
	* 現在のページに、templateから書式をコピーして行を追加します。
	* @param	dst     追加する位置
	* @param	src     書式の行の位置
	* @param	colFr   書式の桁開始位置
	* @param	colTo   書式の桁終了位置
	*/
	public void addRow( int dst, int src, int colFr, int colTo )
		throws Exception
	{
		if ( isDebug()) {
			logger.debug( "copy row " + src + "->"
								+ dst + "[" + colFr + "," + colTo + "]" );
		}
		HSSFRow srcRow = orgSheet.getRow( src );
		HSSFRow dstRow = sheet.createRow( dst );
		dstRow.setHeight( srcRow.getHeight());
		dstRow.setHeightInPoints( srcRow.getHeightInPoints());
		for ( int i = colFr; i < colTo; i++ ) {
			HSSFCell srcCell = srcRow.getCell( i );
			HSSFCell dstCell = dstRow.createCell( i );
			copyCell( dstCell, srcCell );
		}
		Vector<CellRangeAddress> vec = new Vector<CellRangeAddress>();
		for ( int i = 0; i < sheet.getNumMergedRegions(); i++ ) {
			CellRangeAddress reg = sheet.getMergedRegion( i );
			if ( reg.getFirstRow() == src ) {
				reg.setLastRow( dst + reg.getLastRow() - reg.getFirstRow() );
				reg.setFirstRow( dst );
				vec.add( reg );
			}
		}
		for ( int i = 0; i < vec.size(); i++ ) {
			sheet.addMergedRegion( (CellRangeAddress )vec.get( i ) );
		}
	}
	/**
	* 行の処理を終了します。
	*/
	public void endRow() throws Exception {
	}

}
