package	jp.co.sds_corp.report.source;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;

import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
* ファイルからデータを取得するクラスの定義です。
<PRE>
&lt;Source class="jp.co.sds_corp.report.source.FileSource"
	file="src.txt" /&gt;
&lt;Source class="jp.co.sds_corp.report.source.FileSource"
	url="file://src.txt" /&gt;
</PRE>
*/
public class FileSource extends AbstractSource
{
	CmnLogger logger = CmnLogFactory.getInstance().getLogger(this);
	String	file;
	String	url;
	InputStream	stream;
	BufferedReader	reader;

	/**
	* 設定情報で初期化します。
	* @param    config  設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		file = config.getAttributes().get( "file", null );
		url = config.getAttributes().get( "url", null );
		if ( file == null && url == null ) {
			throw( new Exception());
		}
		if ( isDebug()) {
			logger.debug( "file=" + file + " url=" + url );
		}
	}

	/**
	* データを取得します。
	*/
	public void execute() throws Exception {
		if ( file != null ) {
			if ( isDebug()) {
				logger.debug( "open file [" + file + "]" );
			}
			stream = new FileInputStream( file );
		}
		if ( url != null ) {
			if ( isDebug()) {
				logger.debug( "open url [" + url + "]" );
			}
			stream = new URL( url ).openStream();
		}
		reader = new BufferedReader( new InputStreamReader( stream ));
		for ( ;; ) {
			String line = reader.readLine();
			if ( line == null ) {
				break;
			}
			parseLine( line );
		}
		reader.close();
	}

	protected void parseLine( String line ) {
		bean.getList().add( line );
	}

	/**
	* データの取得を終了します。
	*/
	public void exit() {
	}

}
