package	jp.co.sds_corp.report.source;

import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.db.DbAccess;
import jp.co.sds_corp.util.db.DbTable;

/**
* DBからデータを取得するクラスの定義です。
<PRE>
&lt;Source class="jp.co.sds_corp.report.source.SqlSource" &gt;
	&lt;db type="jdbc"&gt;
		&lt;driver&gt;com.sybase.jdbc2.jdbc.SybDriver&lt;/driver&gt;
		&lt;url&gt;jdbc:sybase:Tds:localhost:4100/db&lt;/url&gt;
		&lt;user&gt;user&lt;/user&gt;
		&lt;password&gt;password&lt;/password&gt;
		&lt;encode&gt;MS932&lt;/encode&gt;
	&lt;/db&gt;
	&lt;SQL&gt;
		select * from table where key='A'
	&lt;/SQL&gt;
&lt;/Source&gt;
</PRE>
db要素で接続先を指定します。詳細は{@link DbAccess}を参照してください。
SQL要素で、実行するSQLを指定します。
*/
public class SqlSource extends AbstractSource
{
	DbAccess	db;
	String		sql;

	/**
	* 設定情報で初期化します。
	* @param    config  設定情報
	*/
	public void init( IConfiguration config ) throws Exception {
		super.init( config );
		IConfiguration c = config.getChild( "db" );
		db = new DbAccess( c );
		sql = config.getChild( "SQL" ).get();
		bean = new GenericBean();
	}

	/**
	* データを取得します。
	*/
	public void execute() throws Exception {
		try {
			db.connect();
			db.executeQuery( sql );
			DbTable tbl = new DbTable();
			for ( ;; ) {
				if ( !db.fetch( tbl )) {
					break;
				}
				bean.getList().add( tbl.clone() );
			}
		} finally {
			try {
				db.disconnect();
			} catch ( Exception ex ) {
			}
		}
	}

	/**
	* データの取得を終了します。
	*/
	public void exit() {
	}

}
